def mock_return_once(m, first_result, rest):
    """
    Sets up mock callable such that it returns a result first the time, and a different result all other times it's called.
    Useful for functions that may be called indefinitely in a while loop
    :param m: the mock object
    :param first_result: value to return 1st time the mock is called
    :param rest: value to return after 1st time the mock is called
    :return:
    """

    def f(*args, **kwargs):
        if f.called:
            return rest

        f.called = True
        return first_result

    f.called = False

    m.side_effect = f



import json
import uuid
from unittest import mock

import pytest

try:
    from hedwig.backends.aws import AWSMetadata
except ImportError:
    pass
from hedwig.backends.exceptions import PartialFailure
from hedwig.conf import settings as hedwig_settings
from hedwig.exceptions import ValidationError, CallbackNotFound
from hedwig.testing.factories import MessageFactory

from tests.models import MessageType
from tests.utils import mock_return_once

aws = pytest.importorskip('hedwig.backends.aws')


class TestSNSPublisher:
    def test_publish_success(self, mock_boto3, message):
        sns_publisher = aws.AWSSNSPublisherBackend()
        queue = mock.MagicMock()
        sns_publisher.sns_client.publish.get_queue_by_name = mock.MagicMock(return_value=queue)

        sns_publisher.publish(message)

        mock_boto3.client.assert_called_once_with(
            'sns',
            region_name=hedwig_settings.AWS_REGION,
            aws_access_key_id=hedwig_settings.AWS_ACCESS_KEY,
            aws_secret_access_key=hedwig_settings.AWS_SECRET_KEY,
            aws_session_token=hedwig_settings.AWS_SESSION_TOKEN,
            endpoint_url=hedwig_settings.AWS_ENDPOINT_SQS,
            config=mock.ANY,
        )
        topic = sns_publisher._get_sns_topic(message)
        sns_publisher.sns_client.publish.assert_called_once_with(
            TopicArn=topic,
            Message=sns_publisher.message_payload(message.as_dict()),
            MessageAttributes={k: {'DataType': 'String', 'StringValue': str(v)} for k, v in message.headers.items()},
        )

    @mock.patch('tests.handlers._trip_created_handler', autospec=True)
    def test_sync_mode(self, callback_mock, mock_boto3, message, mock_publisher_backend, settings):
        settings.HEDWIG_PUBLISHER_BACKEND = 'hedwig.backends.aws.AWSSNSPublisherBackend'
        settings.HEDWIG_CONSUMER_BACKEND = 'hedwig.backends.aws.AWSSQSConsumerBackend'
        settings.HEDWIG_SYNC = True

        message.publish()
        callback_mock.assert_called_once_with(message)

    def test_sync_mode_detects_invalid_callback(self, settings, mock_boto3):
        settings.HEDWIG_PUBLISHER_BACKEND = 'hedwig.backends.aws.AWSSNSPublisherBackend'
        settings.HEDWIG_CONSUMER_BACKEND = 'hedwig.backends.aws.AWSSQSConsumerBackend'
        settings.HEDWIG_SYNC = True

        message = MessageFactory(msg_type=MessageType.vehicle_created)
        with pytest.raises(ValidationError) as exc_info:
            message.publish()
        assert isinstance(exc_info.value.__context__, CallbackNotFound)


pre_process_hook = mock.MagicMock()
post_process_hook = mock.MagicMock()


@pytest.fixture(name='sqs_consumer')
def _sqs_consumer(mock_boto3):
    return aws.AWSSQSConsumerBackend()


@pytest.fixture(name='sns_consumer')
def _sns_consumer(mock_boto3):
    return aws.AWSSNSConsumerBackend()


@pytest.fixture(name='prepost_process_hooks')
def _prepost_process_hooks(settings):
    settings.HEDWIG_PRE_PROCESS_HOOK = 'tests.test_backends.test_aws.pre_process_hook'
    settings.HEDWIG_POST_PROCESS_HOOK = 'tests.test_backends.test_aws.post_process_hook'
    yield
    pre_process_hook.reset_mock()
    post_process_hook.reset_mock()


class TestSQSConsumer:
    def test_client_and_resource_instantiation(self, sqs_consumer, mock_boto3):
        # make sure client and resource are initialized
        sqs_consumer.sqs_client
        sqs_consumer.sqs_resource
        mock_boto3.resource.assert_called_once_with(
            'sqs',
            region_name=hedwig_settings.AWS_REGION,
            aws_access_key_id=hedwig_settings.AWS_ACCESS_KEY,
            aws_secret_access_key=hedwig_settings.AWS_SECRET_KEY,
            aws_session_token=hedwig_settings.AWS_SESSION_TOKEN,
            endpoint_url=hedwig_settings.AWS_ENDPOINT_SQS,
        )
        mock_boto3.client.assert_called_once_with(
            'sqs',
            region_name=hedwig_settings.AWS_REGION,
            aws_access_key_id=hedwig_settings.AWS_ACCESS_KEY,
            aws_secret_access_key=hedwig_settings.AWS_SECRET_KEY,
            aws_session_token=hedwig_settings.AWS_SESSION_TOKEN,
            endpoint_url=hedwig_settings.AWS_ENDPOINT_SQS,
        )

    def test_pull_messages(self, sqs_consumer):
        num_messages = 1
        visibility_timeout = 10
        queue = mock.MagicMock()
        sqs_consumer.sqs_resource.get_queue_by_name = mock.MagicMock(return_value=queue)

        sqs_consumer.pull_messages(num_messages, visibility_timeout)

        sqs_consumer.sqs_resource.get_queue_by_name.assert_called_once_with(QueueName=sqs_consumer.queue_name)
        queue.receive_messages.assert_called_once_with(
            MaxNumberOfMessages=num_messages,
            MessageAttributeNames=['All'],
            VisibilityTimeout=visibility_timeout,
            WaitTimeSeconds=sqs_consumer.WAIT_TIME_SECONDS,
        )

    def test_extend_visibility_timeout(self, sqs_consumer):
        visibility_timeout_s = 10
        receipt = "receipt"
        sqs_consumer.sqs_client.get_queue_url = mock.MagicMock(return_value={"QueueUrl": "DummyQueueUrl"})

        sqs_consumer.extend_visibility_timeout(visibility_timeout_s, AWSMetadata(receipt))

        sqs_consumer.sqs_client.get_queue_url.assert_called_once_with(QueueName=sqs_consumer.queue_name)
        sqs_consumer.sqs_client.change_message_visibility.assert_called_once_with(
            QueueUrl='DummyQueueUrl', ReceiptHandle='receipt', VisibilityTimeout=10
        )

    def test_success_requeue_dead_letter(self, sqs_consumer):
        sqs_consumer = aws.AWSSQSConsumerBackend(dlq=True)
        num_messages = 3
        visibility_timeout = 4

        messages = [mock.MagicMock() for _ in range(num_messages)]
        sqs_consumer.pull_messages = mock.MagicMock(side_effect=iter([messages, []]))

        mock_queue, mock_dlq = mock.MagicMock(), mock.MagicMock()
        mock_queue.send_messages.return_value = {'Failed': []}
        sqs_consumer.sqs_resource.get_queue_by_name = mock.MagicMock(side_effect=iter([mock_queue, mock_dlq]))
        mock_dlq.delete_messages.return_value = {'Failed': []}

        sqs_consumer.requeue_dead_letter(num_messages=num_messages, visibility_timeout=visibility_timeout)

        sqs_consumer.sqs_resource.get_queue_by_name.assert_has_calls(
            [
                mock.call(QueueName=f'HEDWIG-{hedwig_settings.HEDWIG_QUEUE}'),
                mock.call(QueueName=sqs_consumer.queue_name),
            ]
        )

        sqs_consumer.pull_messages.assert_has_calls(
            [
                mock.call(num_messages=num_messages, visibility_timeout=visibility_timeout),
                mock.call(num_messages=num_messages, visibility_timeout=visibility_timeout),
            ]
        )
        mock_queue.send_messages.assert_called_once_with(
            Entries=[
                {
                    'Id': queue_message.message_id,
                    'MessageBody': queue_message.body,
                    'MessageAttributes': queue_message.message_attributes,
                }
                for queue_message in messages
            ]
        )
        mock_dlq.delete_messages.assert_called_once_with(
            Entries=[
                {'Id': queue_message.message_id, 'ReceiptHandle': queue_message.receipt_handle}
                for queue_message in messages
            ]
        )

    def test_partial_failure_requeue_dead_letter(self, sqs_consumer):
        num_messages = 1
        visibility_timeout = 4
        queue_name = "HEDWIG-DEV-RTEP"

        messages = [mock.MagicMock() for _ in range(num_messages)]
        sqs_consumer.pull_messages = mock.MagicMock(side_effect=iter([messages, None]))
        dlq_name = f'{queue_name}-DLQ'

        mock_queue, mock_dlq = mock.MagicMock(), mock.MagicMock()
        mock_queue.attributes = {'RedrivePolicy': json.dumps({'deadLetterTargetArn': dlq_name})}
        mock_queue.send_messages.return_value = {'Successful': ['success_id'], 'Failed': ["fail_id"]}
        sqs_consumer._get_queue_by_name = mock.MagicMock(side_effect=iter([mock_queue, mock_dlq]))

        with pytest.raises(PartialFailure):
            sqs_consumer.requeue_dead_letter(num_messages=num_messages, visibility_timeout=visibility_timeout)

    def test_fetch_and_process_messages_success(
        self, sqs_consumer, message_data, timed_shutdown_event, prepost_process_hooks
    ):
        num_messages = 3
        visibility_timeout = 4
        queue = mock.MagicMock()
        sqs_consumer.sqs_resource.get_queue_by_name = mock.MagicMock(return_value=queue)

        queue_message = mock.MagicMock()
        queue_message.body = json.dumps(message_data)
        queue_message.receipt_handle = "dummy receipt"

        mock_return_once(queue.receive_messages, [queue_message], [])
        message_mock = mock.MagicMock()
        sqs_consumer._build_message = mock.MagicMock(return_value=message_mock)
        sqs_consumer.process_message = mock.MagicMock(wraps=sqs_consumer.process_message)
        sqs_consumer.message_handler = mock.MagicMock(wraps=sqs_consumer.message_handler)

        sqs_consumer.fetch_and_process_messages(num_messages, visibility_timeout, timed_shutdown_event)

        sqs_consumer.sqs_resource.get_queue_by_name.assert_called_with(QueueName=sqs_consumer.queue_name)
        queue.receive_messages.assert_called_with(
            MaxNumberOfMessages=num_messages,
            MessageAttributeNames=['All'],
            VisibilityTimeout=visibility_timeout,
            WaitTimeSeconds=sqs_consumer.WAIT_TIME_SECONDS,
        )
        sqs_consumer.process_message.assert_called_once_with(queue_message)
        sqs_consumer.message_handler.assert_called_once_with(
            queue_message.body, AWSMetadata(queue_message.receipt_handle)
        )
        message_mock.exec_callback.assert_called_once_with()
        queue_message.delete.assert_called_once_with()
        pre_process_hook.assert_called_once_with(sqs_queue_message=queue_message)
        post_process_hook.assert_called_once_with(sqs_queue_message=queue_message)


class TestSNSConsumer:
    @mock.patch('hedwig.backends.aws.AWSSNSConsumerBackend.process_message')
    def test_process_messages(self, mock_process_message, sns_consumer):
        records = mock.Mock(), mock.Mock()
        event = {'Records': records}

        sns_consumer.process_messages(event)

        mock_process_message.assert_has_calls([mock.call(r) for r in records])

    def test_success_process_message(self, sns_consumer, prepost_process_hooks):
        # copy from https://docs.aws.amazon.com/lambda/latest/dg/eventsources.html#eventsources-sns
        mock_record = {
            "EventVersion": "1.0",
            "EventSubscriptionArn": "arn",
            "EventSource": "aws:sns",
            "Sns": {
                "SignatureVersion": "1",
                "Timestamp": "1970-01-01T00:00:00.000Z",
                "Signature": "EXAMPLE",
                "SigningCertUrl": "EXAMPLE",
                "MessageId": "95df01b4-ee98-5cb9-9903-4c221d41eb5e",
                "Message": "Hello from SNS!",
                "MessageAttributes": {
                    "request_id": {"Type": "String", "Value": str(uuid.uuid4())},
                    "TestBinary": {"Type": "Binary", "Value": "TestBinary"},
                },
                "Type": "Notification",
                "UnsubscribeUrl": "EXAMPLE",
                "TopicArn": "arn",
                "Subject": "TestInvoke",
            },
        }
        message_mock = mock.MagicMock()
        sns_consumer._build_message = mock.MagicMock(return_value=message_mock)

        sns_consumer.process_message(mock_record)

        pre_process_hook.assert_called_once_with(sns_record=mock_record)
        post_process_hook.assert_called_once_with(sns_record=mock_record)
        message_mock.exec_callback.assert_called_once_with()



import logging
from contextlib import contextmanager
from unittest import mock

import pytest

import hedwig.conf
from hedwig import models
from hedwig.backends.base import HedwigBaseBackend, HedwigPublisherBaseBackend
from hedwig.backends.utils import get_publisher_backend, get_consumer_backend
from hedwig.models import Message
from hedwig.testing.factories import MessageFactory

from tests.models import MessageType

try:
    # may not be available
    from moto import mock_sqs, mock_sns
except ImportError:
    pass


def pytest_configure():
    logging.basicConfig()


@pytest.fixture
def settings():
    """
    Use this fixture to override settings. Changes are automatically reverted
    """
    original_module = hedwig.conf.settings._user_settings

    class Wrapped:
        # default to the original module, but allow tests to setattr which would override
        def __getattr__(self, name):
            return getattr(original_module, name)

    hedwig.conf.settings._user_settings = Wrapped()
    hedwig.conf.settings.clear_cache()

    # since consumer/publisher settings may have changed
    get_publisher_backend.cache_clear()
    get_consumer_backend.cache_clear()

    # in case a test overrides HEDWIG_DATA_VALIDATOR_CLASS
    models._validator = None

    try:
        yield hedwig.conf.settings._user_settings
    finally:
        hedwig.conf.settings._user_settings = original_module
        hedwig.conf.settings.clear_cache()

        # in case a test overrides HEDWIG_DATA_VALIDATOR_CLASS
        models._validator = None

        # since consumer/publisher settings may have changed
        get_publisher_backend.cache_clear()
        get_consumer_backend.cache_clear()


@pytest.fixture(name='message_data')
def _message_data():
    return MessageFactory.build(msg_type=MessageType.trip_created)


@pytest.fixture()
def message(message_data):
    return Message(message_data)


@contextmanager
def _mock_boto3():
    settings.AWS_REGION = 'us-west-1'
    with mock_sqs(), mock_sns(), mock.patch("hedwig.backends.aws.boto3", autospec=True) as boto3_mock:
        yield boto3_mock


@pytest.fixture
def mock_boto3():
    with _mock_boto3() as m:
        yield m


@pytest.fixture()
def sqs_consumer_backend(mock_boto3):
    # may not be available
    from hedwig.backends import aws

    yield aws.AWSSQSConsumerBackend()


@pytest.fixture
def mock_pubsub_v1():
    with mock.patch("hedwig.backends.gcp.pubsub_v1", autospec=True) as pubsub_v1_mock:
        yield pubsub_v1_mock


@pytest.fixture(params=['aws', 'google'])
def consumer_backend(request):
    if request.param == 'aws':
        try:
            import hedwig.backends.aws  # noqa

            with _mock_boto3():
                yield HedwigBaseBackend.build("hedwig.backends.aws.AWSSQSConsumerBackend")
        except ImportError:
            pytest.skip("AWS backend not importable")

    if request.param == 'google':
        try:
            import hedwig.backends.gcp  # noqa

            with mock.patch("hedwig.backends.gcp.pubsub_v1"), mock.patch(
                "hedwig.backends.gcp.google_auth_default", return_value=(None, "DUMMY")
            ):
                yield HedwigBaseBackend.build("hedwig.backends.gcp.GooglePubSubConsumerBackend")
        except ImportError:
            pytest.skip("Google backend not importable")


@pytest.fixture(
    params=["hedwig.backends.aws.AWSSNSConsumerBackend", "hedwig.backends.gcp.GooglePubSubPublisherBackend"]
)
def publisher_backend(request, mock_boto3):
    with mock.patch("hedwig.backends.gcp.pubsub_v1"):
        yield HedwigBaseBackend.build(request.param)


@pytest.fixture()
def mock_publisher_backend():
    with mock.patch.object(HedwigPublisherBaseBackend, '_publish'):
        yield HedwigPublisherBaseBackend()



import importlib


def import_class(dotted_path):
    if not dotted_path:
        raise ImportError(f"{dotted_path} is not defined")
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError as err:
        raise ImportError(f"{dotted_path} doesn't look like a module path") from err

    module = importlib.import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as err:
        raise ImportError(f"Module '{module_path}' does not define a '{class_name}' attribute/class") from err



import json
import logging
import threading
from typing import cast, Mapping, Optional, Generator, List, Union
from unittest import mock

import boto3
import funcy
from botocore.config import Config
from retrying import retry

from hedwig.backends.base import HedwigConsumerBaseBackend, HedwigPublisherBaseBackend
from hedwig.backends.exceptions import PartialFailure
from hedwig.conf import settings
from hedwig.models import Message


logger = logging.getLogger(__name__)


class AWSMetadata:
    """
    AWS specific metadata for a Message
    """

    def __init__(self, receipt: str):
        self._receipt: str = receipt

    @property
    def receipt(self) -> str:
        """
        AWS receipt identifier
        """
        return self._receipt

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, AWSMetadata):
            return False
        return self._receipt == o._receipt

    def __ne__(self, o: object) -> bool:
        return not self.__eq__(o)

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return f'AWSMetadata(receipt={self._receipt})'

    def __hash__(self) -> int:
        return hash((self._receipt,))


class AWSSNSPublisherBackend(HedwigPublisherBaseBackend):
    def __init__(self):
        self._sns_client = None

    @property
    def sns_client(self):
        if self._sns_client is None:
            config = Config(connect_timeout=settings.AWS_CONNECT_TIMEOUT_S, read_timeout=settings.AWS_READ_TIMEOUT_S)
            self._sns_client = boto3.client(
                'sns',
                region_name=settings.AWS_REGION,
                aws_access_key_id=settings.AWS_ACCESS_KEY,
                aws_secret_access_key=settings.AWS_SECRET_KEY,
                aws_session_token=settings.AWS_SESSION_TOKEN,
                endpoint_url=settings.AWS_ENDPOINT_SNS,
                config=config,
            )
        return self._sns_client

    @staticmethod
    def _get_sns_topic(message: Message) -> str:
        return f'arn:aws:sns:{settings.AWS_REGION}:{settings.AWS_ACCOUNT_ID}:hedwig-{message.topic}'

    @retry(stop_max_attempt_number=3, stop_max_delay=3000)
    def _publish_over_sns(self, topic: str, message_json: str, message_attributes: Mapping) -> str:
        # transform (http://boto.cloudhackers.com/en/latest/ref/sns.html#boto.sns.SNSConnection.publish)
        message_attributes = {k: {'DataType': 'String', 'StringValue': str(v)} for k, v in message_attributes.items()}
        response = self.sns_client.publish(TopicArn=topic, Message=message_json, MessageAttributes=message_attributes)
        return response['PublishResponse']['PublishResult']['MessageId']

    def _mock_queue_message(self, message: Message) -> mock.Mock:
        sqs_message = mock.Mock()
        sqs_message.body = json.dumps(message.as_dict())
        sqs_message.receipt_handle = 'test-receipt'
        return sqs_message

    def _publish(self, message: Message, payload: str, headers: Optional[Mapping] = None) -> str:
        topic = self._get_sns_topic(message)
        return self._publish_over_sns(topic, payload, headers)


class AWSSQSConsumerBackend(HedwigConsumerBaseBackend):
    WAIT_TIME_SECONDS = 20

    def __init__(self, dlq=False):
        self._sqs_resource = None
        self._sqs_client = None
        self.queue_name = f'HEDWIG-{settings.HEDWIG_QUEUE}{"-DLQ" if dlq else ""}'

    @property
    def sqs_resource(self):
        if self._sqs_resource is None:
            self._sqs_resource = boto3.resource(
                'sqs',
                region_name=settings.AWS_REGION,
                aws_access_key_id=settings.AWS_ACCESS_KEY,
                aws_secret_access_key=settings.AWS_SECRET_KEY,
                aws_session_token=settings.AWS_SESSION_TOKEN,
                endpoint_url=settings.AWS_ENDPOINT_SQS,
            )
        return self._sqs_resource

    @property
    def sqs_client(self):
        if self._sqs_client is None:
            self._sqs_client = boto3.client(
                'sqs',
                region_name=settings.AWS_REGION,
                aws_access_key_id=settings.AWS_ACCESS_KEY,
                aws_secret_access_key=settings.AWS_SECRET_KEY,
                aws_session_token=settings.AWS_SESSION_TOKEN,
                endpoint_url=settings.AWS_ENDPOINT_SQS,
            )
        return self._sqs_client

    def _get_queue(self):
        return self.sqs_resource.get_queue_by_name(QueueName=self.queue_name)

    def pull_messages(
        self, num_messages: int = 10, visibility_timeout: int = None, shutdown_event: Optional[threading.Event] = None
    ) -> Union[Generator, List]:
        """

        :param shutdown_event: Unused for this backend
        :param num_messages:
        :param visibility_timeout:
        :return:
        """
        params = {
            'MaxNumberOfMessages': num_messages,
            'WaitTimeSeconds': self.WAIT_TIME_SECONDS,
            'MessageAttributeNames': ['All'],
        }
        if visibility_timeout is not None:
            params['VisibilityTimeout'] = visibility_timeout
        return self._get_queue().receive_messages(**params)

    def process_message(self, queue_message) -> None:
        message_json = queue_message.body
        receipt = queue_message.receipt_handle
        self.message_handler(message_json, AWSMetadata(receipt))

    def ack_message(self, queue_message) -> None:
        queue_message.delete()

    def nack_message(self, queue_message) -> None:
        # let visibility timeout take care of it
        pass

    def extend_visibility_timeout(self, visibility_timeout_s: int, metadata: AWSMetadata) -> None:
        """
        Extends visibility timeout of a message on a given priority queue for long running tasks.
        """
        receipt = metadata.receipt
        queue_url = self.sqs_client.get_queue_url(QueueName=self.queue_name)['QueueUrl']
        self.sqs_client.change_message_visibility(
            QueueUrl=queue_url, ReceiptHandle=receipt, VisibilityTimeout=visibility_timeout_s
        )

    def requeue_dead_letter(self, num_messages: int = 10, visibility_timeout: int = None) -> None:
        """
        Re-queues everything in the Hedwig DLQ back into the Hedwig queue.

        :param num_messages: Maximum number of messages to fetch in one SQS call. Defaults to 10.
        :param visibility_timeout: The number of seconds the message should remain invisible to other queue readers.
        Defaults to None, which is queue default
        """
        sqs_queue = self.sqs_resource.get_queue_by_name(QueueName=f'HEDWIG-{settings.HEDWIG_QUEUE}')
        dead_letter_queue = self._get_queue()

        logging.info("Re-queueing messages from {} to {}".format(dead_letter_queue.url, sqs_queue.url))
        while True:
            queue_messages = self.pull_messages(num_messages=num_messages, visibility_timeout=visibility_timeout)
            queue_messages = cast(list, queue_messages)

            if not queue_messages:
                break

            logging.info("got {} messages from dlq".format(len(queue_messages)))

            result = sqs_queue.send_messages(
                Entries=[
                    funcy.merge(
                        {'Id': queue_message.message_id, 'MessageBody': queue_message.body},
                        {'MessageAttributes': queue_message.message_attributes}
                        if queue_message.message_attributes
                        else {},
                    )
                    for queue_message in queue_messages
                ]
            )
            if result.get('Failed'):
                raise PartialFailure(result)

            dead_letter_queue.delete_messages(
                Entries=[
                    {'Id': message.message_id, 'ReceiptHandle': message.receipt_handle} for message in queue_messages
                ]
            )

            logging.info("Re-queued {} messages".format(len(queue_messages)))

    @staticmethod
    def pre_process_hook_kwargs(queue_message) -> dict:
        return {'sqs_queue_message': queue_message}

    @staticmethod
    def post_process_hook_kwargs(queue_message) -> dict:
        return {'sqs_queue_message': queue_message}


class AWSSNSConsumerBackend(HedwigConsumerBaseBackend):
    def requeue_dead_letter(self, num_messages: int = 10, visibility_timeout: int = None) -> None:
        raise RuntimeError("invalid operation for backend")

    def pull_messages(
        self, num_messages: int = 10, visibility_timeout: int = None, shutdown_event: Optional[threading.Event] = None
    ) -> Union[Generator, List]:
        raise RuntimeError("invalid operation for backend")

    def ack_message(self, queue_message) -> None:
        raise RuntimeError("invalid operation for backend")

    def nack_message(self, queue_message) -> None:
        raise RuntimeError("invalid operation for backend")

    def extend_visibility_timeout(self, visibility_timeout_s: int, metadata) -> None:
        raise RuntimeError("invalid operation for backend")

    def process_messages(self, lambda_event):
        for record in lambda_event['Records']:
            self.process_message(record)

    def process_message(self, queue_message) -> None:
        settings.HEDWIG_PRE_PROCESS_HOOK(sns_record=queue_message)
        message_json = queue_message['Sns']['Message']
        self.message_handler(message_json, None)
        settings.HEDWIG_POST_PROCESS_HOOK(sns_record=queue_message)



import os
from contextlib import contextmanager
from functools import lru_cache
from typing import Any, Generator

from hedwig.conf import settings


@contextmanager
def override_env(env: str, value: Any) -> Generator[None, None, None]:
    """
    Override environment variable value temporarily
    """
    orig_value = os.environ.get(env)

    os.environ[env] = value

    try:
        yield
    finally:
        # was the value originally set? if so, restore it
        if orig_value is not None:
            os.environ[env] = orig_value
        else:
            # value wasn't set originally, so unset it again
            del os.environ[env]


@lru_cache(maxsize=3)
def get_publisher_backend(*args, **kwargs):
    from hedwig.backends.base import HedwigPublisherBaseBackend

    return HedwigPublisherBaseBackend.build(settings.HEDWIG_PUBLISHER_BACKEND, *args, **kwargs)


@lru_cache(maxsize=3)
def get_consumer_backend(*args, **kwargs):
    from hedwig.backends.base import HedwigConsumerBaseBackend

    return HedwigConsumerBaseBackend.build(settings.HEDWIG_CONSUMER_BACKEND, *args, **kwargs)



from hedwig.backends.utils import get_consumer_backend


def requeue_dead_letter(num_messages: int = 10, visibility_timeout: int = None) -> None:
    """
    Re-queues everything in the Hedwig DLQ back into the Hedwig queue.

    :param num_messages: Maximum number of messages to fetch in one call. Defaults to 10.
    :param visibility_timeout: The number of seconds the message should remain invisible to other queue readers.
        Defaults to None, which is queue default
    """
    consumer_backend = get_consumer_backend(dlq=True)
    consumer_backend.requeue_dead_letter(num_messages, visibility_timeout)
